/*
 * InstanceReader.cpp
 *
 *  Created on: 8 Mar 2015
 *      Author: mruthmair
 */

#include <boost/algorithm/string.hpp>
#include <boost/tokenizer.hpp>
#include <string>
#include <fstream>
#include <vector>

int InstanceReader::readFile_GLR( InputGraph& g )
{
	ifstream file( params.ipathfile.c_str() );
	if( !file.good() ) {
		cout << "File " << params.ipathfile << " does not exist.\n";
		return -1;
	}
	cout << "Importing TSP instance from GLR file " << params.ipathfile << "\n";

	// number of nodes is given, the first <nnodes> are taken from the file
	g.addVertices( params.nnodes, InputGraph::WHITE );

	// line 1: number of vertices (should be 200)
	u_int n_vertices;
	file >> n_vertices;
	assert( n_vertices == 200 );
	vector<u_int> v2i( n_vertices, NODE_NOT_EXISTS ); // mapping from input file node index to node index used here

	// line 2 - (<n_vertices> + 1): node coordinates
	struct NodeT
	{
		int x, y;
		u_int i;
	};
	list<NodeT> points;

	for( u_int i = 0; i < n_vertices; i++ ) {
		NodeT v;
		file >> v.x >> v.y;
		v.i = i;
		if( i < g.n_vertices ) points.push_back( v );
	}

	// re-sort nodes according to distribution of blacks (blacks are always the first nodes in the list)
	switch( params.distribution_of_blacks ) {
		case BD_RND:
		    // random set of black nodes -> the list is not re-ordered
			break;
		case BD_CENTER:
		    // nodes are ordered by increasing Euclidian distance to the center
			points.sort( [&g](const NodeT& p1, const NodeT& p2) -> bool {return (p1.x - 500) * (p1.x - 500) + (p1.y - 500) * (p1.y - 500) < (p2.x - 500) * (p2.x - 500) + (p2.y - 500) * (p2.y - 500);} );
			break;
		case BD_CORNER:
		    // nodes are ordered by increasing Manhattan distance to the (0,0)-corner
			points.sort( [&g](const NodeT& p1, const NodeT& p2) -> bool {return p1.x + p1.y < p2.x + p2.y;} );
			break;
		case BD_GEOCLUST:
			// first node is fixed to be a black, further blacks should have maximal distance to all previous blacks
			list<NodeT> sorted;
			sorted.splice( sorted.begin(), points, points.begin() );
			while( !points.empty() ) {
				auto p_maxdist = points.begin();
				int maxdist = 0;
				for( auto pi = points.begin(); pi != points.end(); pi++ ) {
					int mindist = INT_MAX;
					for( auto& b : sorted ) {
						int dx = pi->x - b.x;
						int dy = pi->y - b.y;
						int dist = dx * dx + dy * dy;
						if( dist < mindist ) mindist = dist;
					}
					if( mindist > maxdist ) {
						maxdist = mindist;
						p_maxdist = pi;
					}
				}
				sorted.splice( sorted.end(), points, p_maxdist );
			}
			points.splice( points.begin(), sorted );
			break;
	}

	// set node indices and coordinates
	u_int i = 0;
	for( auto& p : points ) {
		g.setCoordinates( i, make_pair( p.x, p.y ) );
		v2i[p.i] = i;
		i++;
	}

	// add edges and calculate costs
	for( u_int i = 0; i < g.n_vertices; i++ ) {
		for( u_int j = i + 1; j < g.n_vertices; j++ ) {
			double dx = (double) g.getX( i ) - (double) g.getX( j );
			double dy = (double) g.getY( i ) - (double) g.getY( j );
			CostT c = floor( sqrt( dx * dx + dy * dy ) + 0.5 );
			g.addEdge( i, j, c );
		}
	}

	// read distances from file
	g.distance_equal_to_cost = (params.ubdist == 0);
	for( u_int u = 0; u < g.n_vertices; u++ ) {
		for( u_int v = 0; v < u; v++ ) {
			// take care about original node indices
			u_int i = v2i[u], j = v2i[v];
			if( i > j ) swap( i, j );
			DistanceT d;
			if( params.ubdist > 0 ) file >> d;
			else d = g.getEdgeCost( i, j );
			g.setEdgeDistance( i, j, d );
		}
	}
	file.close();

	// set "virtual" start node
	g.start = 0;
	// hop limit: defined by alpha
	g.Q = round( params.alpha * (float) g.n_vertices );
	// set of black nodes
	u_int bmin = 0;
	while( bmin - ceil( (float) (g.n_vertices - bmin) / (float) g.Q ) < 0 )
		bmin++;
	u_int n_blacks = min( g.n_vertices, (u_int) ceil( (float) bmin * params.beta ) );
	for( u_int i = 0; i < n_blacks; i++ )
		g.setColor( i, InputGraph::BLACK );

	// path length limit
	g.L = getDistanceLimitFromFile( "Lbounds.csv", g.n_vertices, params.inr, params.ubdist, params.distribution_of_blacks, g.Q, g.n_blacks );

	cout << "|B| = " << g.n_blacks << ", bmin = " << bmin << ", beta = " << params.beta << "\n";
	cout << "Q = " << g.Q << ", alpha = " << params.alpha << "\n";

	cout << "Importing finished.\n";
	return 0;
}

int InstanceReader::readFile_MUT( InputGraph& g )
{
	ifstream file( params.ipathfile.c_str() );
	if( !file.good() ) {
		cout << "File " << params.ipathfile << " does not exist.\n";
		return -1;
	}
	cout << "Importing TSP instance from Muter-2015 file " << params.ipathfile << "\n";

	// number of nodes is given, the first <nnodes> are taken from the file
	g.addVertices( params.nnodes, InputGraph::WHITE );

	// line 1: Q, bmin, Lmax
	u_int bmin;
	double Lmax; // IS NOT RELIABLE -> TAKE FROM LBOUNDS-FILE!
	file >> g.Q >> bmin >> Lmax;

	// line 2 - EOF: node coordinates
	for( u_int i = 0; i < g.n_vertices; i++ ) {
		pair<int, int> coords;
		file >> coords.first >> coords.second;
		g.setCoordinates( i, coords );
	}
	file.close();

	// calculate arc costs (equal to distance if UBDIST=0)
	g.distance_equal_to_cost = (params.ubdist == 0);
	for( u_int i = 0; i < g.n_vertices; i++ ) {
		for( u_int j = i + 1; j < g.n_vertices; j++ ) {
			double dx = (double) g.getX( i ) - (double) g.getX( j );
			double dy = (double) g.getY( i ) - (double) g.getY( j );
			CostT c = floor( sqrt( dx * dx + dy * dy ) + 0.5 );
			DistanceT d = c;
			g.addEdge( i, j, c, d );
		}
	}

	// set "virtual" start node
	g.start = 0;
	// set of black nodes
	u_int n_blacks = ceil( (float) bmin * params.beta );
	for( u_int i = 0; i < n_blacks; i++ )
		g.setColor( i, InputGraph::BLACK );
	// path length limit
	g.L = getDistanceLimitFromFile( "Lbounds.csv", g.n_vertices, params.inr, params.ubdist, 0, g.Q, g.n_blacks );

	cout << "Q = " << g.Q << ", alpha = " << params.alpha << "\n";
	cout << "bmin = " << bmin << ", beta = " << params.beta << ", |B| = " << g.n_blacks << "\n";

	cout << "Importing finished.\n";
	return 0;
}

